@@ -1,6 +1,7 @@
 use crate::boolean_expression::BooleanExpression;
 use crate::boolean_expression::BooleanExpression::Variable;
 use crate::*;
+use num_bigint::BigInt;
 use std::iter::Map;
 use std::ops::Range;
 use std::slice::Iter;
@@ -28,6 +29,47 @@ impl Bdd {
         self.0.len() == 1
     }
 
+    pub fn exact_cardinality(&self) -> BigInt {
+        let zero = BigInt::from(0);
+        let one = BigInt::from(1);
+        if self.is_false() {
+            return zero;
+        }
+        let mut cache = vec![None; self.0.len()];
+        cache[0] = Some(zero);
+        cache[1] = Some(one.clone());
+        let mut stack: Vec<BddPointer> = vec![self.root_pointer()];
+        while let Some(node) = stack.last() {
+            if cache[node.0 as usize].is_some() {
+                stack.pop();
+            } else {
+                let low = self.low_link_of(*node);
+                let high = self.high_link_of(*node);
+                let low_var = self.var_of(low).0;
+                let high_var = self.var_of(high).0;
+                let node_var = self.var_of(*node).0;
+                let low = low.0 as usize;
+                let high = high.0 as usize;
+
+                if let (Some(cache_low), Some(cache_high)) = (&cache[low], &cache[high]) {
+                    let low_card = cache_low * (one.clone() << (low_var - node_var - 1));
+                    let high_card = cache_high * (one.clone() << (high_var - node_var - 1));
+                    cache[node.0 as usize] = Some(low_card + high_card);
+                    stack.pop();
+                } else {
+                    if cache[low].is_none() {
+                        stack.push(BddPointer(low as u32));
+                    }
+                    if cache[high].is_none() {
+                        stack.push(BddPointer(high as u32));
+                    }
+                }
+            }
+        }
+        let last_var = self.0.last().unwrap().var.0;
+        cache.last().cloned().flatten().unwrap() * (one << last_var)
+    }
+
     /// Approximately computes the number of valuations satisfying the formula given
     /// by this `Bdd`.
     pub fn cardinality(&self) -> f64 {
@@ -127,9 +169,9 @@ impl Bdd {
             let expression = if low_link.is_terminal() && high_link.is_terminal() {
                 // Both links are terminal, which means this is an exactly determined variable
                 if high_link.is_one() && low_link.is_zero() {
-                    BooleanExpression::Variable(var_name)
+                    Variable(var_name)
                 } else if high_link.is_zero() && low_link.is_one() {
-                    BooleanExpression::Not(Box::new(BooleanExpression::Variable(var_name)))
+                    BooleanExpression::Not(Box::new(Variable(var_name)))
                 } else {
                     panic!("Invalid node {:?} in bdd {:?}.", self.0[node], self.0);
                 }
@@ -137,15 +179,13 @@ impl Bdd {
                 if low_link.is_zero() {
                     // a & high
                     BooleanExpression::And(
-                        Box::new(BooleanExpression::Variable(var_name)),
+                        Box::new(Variable(var_name)),
                         Box::new(results[high_link.0 as usize].clone()),
                     )
                 } else {
                     // !a | high
                     BooleanExpression::Or(
-                        Box::new(BooleanExpression::Not(Box::new(
-                            BooleanExpression::Variable(var_name),
-                        ))),
+                        Box::new(BooleanExpression::Not(Box::new(Variable(var_name)))),
                         Box::new(results[high_link.0 as usize].clone()),
                     )
                 }
@@ -153,15 +193,13 @@ impl Bdd {
                 if high_link.is_zero() {
                     // !a & low
                     BooleanExpression::And(
-                        Box::new(BooleanExpression::Not(Box::new(
-                            BooleanExpression::Variable(var_name),
-                        ))),
+                        Box::new(BooleanExpression::Not(Box::new(Variable(var_name)))),
                         Box::new(results[low_link.0 as usize].clone()),
                     )
                 } else {
                     // a | low
                     BooleanExpression::Or(
-                        Box::new(BooleanExpression::Variable(var_name)),
+                        Box::new(Variable(var_name)),
                         Box::new(results[low_link.0 as usize].clone()),
                     )
                 }
@@ -323,6 +361,7 @@ mod tests {
     use crate::_test_util::mk_small_test_bdd;
     use crate::boolean_expression::BooleanExpression;
     use crate::*;
+    use num_bigint::BigInt;
     use std::convert::TryFrom;
 
     #[test]
@@ -359,6 +398,13 @@ mod tests {
         assert_eq!(8.0, bdd.cardinality());
     }
 
+    #[test]
+    fn bdd_exact_cardinality() {
+        // 5 variables, v3 & !v4
+        let bdd = mk_small_test_bdd();
+        assert_eq!(BigInt::from(8), bdd.exact_cardinality());
+    }
+
     #[test]
     fn bdd_sat_witness_basic() {
         // v3 & !v4
